from typing import Any, Collection, Dict, Iterable, Iterator, List, Set, Tuple

from . import annotations as annotations
from .annotations import Gene as Gene
from .annotations import Omim as Omim
from .annotations import Orpha as Orpha
from .types import InformationContent

from . import helper as helper

__version__: str
__backend__: str


class HPOTerm:
    id: str
    name: str
    information_content: InformationContent
    parents: Set[HPOTerm]
    all_parents: Set[HPOTerm]
    children: Set[HPOTerm]
    genes: Set[Gene]
    omim_diseases: Set[Omim]
    orpha_diseases: Set[Orpha]
    categories: List[HPOTerm]
    def parent_of(self, other: HPOTerm) ->  bool: ...
    def child_of(self, other: HPOTerm) -> bool: ...
    def parent_ids(self) -> List[int]: ...
    def common_ancestors(self, other: HPOTerm) -> Set[HPOTerm]: ...
    def count_parents(self) -> int: ...
    def shortest_path_to_root(self) -> int: ...
    def shortest_path_to_parent(self, other: HPOTerm) -> Tuple[float, List[HPOTerm]]: ...
    def path_to_other(self, other: HPOTerm) -> Tuple[int, List[HPOTerm], int, int]: ...
    def similarity_score(self, other: HPOTerm, kind: str = "omim", method: str = "graphic") -> float: ...
    def similarity_scores(self, other: List[HPOTerm], kind: str = "omim", method: str = "graphic") -> List[float]: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __hash__(self) -> int: ...


class HPOSet:
    def __init__(self, terms: List[int | HPOTerm]): ...
    def add(self, term: int | HPOTerm): ...
    def child_nodes(self) -> HPOSet: ...
    def remove_modifier(self) -> HPOSet: ...
    def replace_obsolete(self) -> HPOSet: ...
    def all_genes(self) -> Set[Gene]: ...
    def omim_diseases(self) -> Set[Omim]: ...
    def orpha_diseases(self) -> Set[Orpha]: ...
    def information_content(self) -> Dict[str, Any]: ...
    def similarity(self, other: HPOSet, kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> float: ...
    def similarity_scores(self, other: List[HPOSet], kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> List[float]: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    def serialize(self) -> str: ...
    def to_bytes(self) -> bytes: ...
    def terms(self) -> Iterator[HPOTerm]: ...
    @classmethod
    def from_queries(cls, queries: List[int | str]) -> HPOSet: ...
    @classmethod
    def from_serialized(cls, pickle: str) -> HPOSet: ...
    @classmethod
    def from_bytes(cls, bytes: bytes) -> HPOSet: ...
    @classmethod
    def from_gene(cls, gene: Gene) -> HPOSet: ...
    @classmethod
    def from_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_omim_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_orpha_disease(cls, disease: Orpha) -> HPOSet: ...
    def __len__(self) -> int: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> Iterator[HPOTerm]: ...
    def __contains__(self, term: HPOTerm) -> bool: ...


class BasicHPOSet:
    def __init__(self, terms: List[int | HPOTerm]): ...
    def add(self, term: int | HPOTerm): ...
    def child_nodes(self) -> HPOSet: ...
    def remove_modifier(self) -> HPOSet: ...
    def replace_obsolete(self) -> HPOSet: ...
    def all_genes(self) -> Set[Gene]: ...
    def omim_diseases(self) -> Set[Omim]: ...
    def orpha_diseases(self) -> Set[Orpha]: ...
    def information_content(self) -> Dict[str, Any]: ...
    def similarity(self, other: HPOSet, kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> float: ...
    def similarity_scores(self, other: List[HPOSet], kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> List[float]: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    def serialize(self) -> str: ...
    def to_bytes(self) -> bytes: ...
    def terms(self) -> Iterator[HPOTerm]: ...
    @classmethod
    def from_queries(cls, queries: List[int | str]) -> HPOSet: ...
    @classmethod
    def from_serialized(cls, pickle: str) -> HPOSet: ...
    @classmethod
    def from_bytes(cls, bytes: bytes) -> HPOSet: ...
    @classmethod
    def from_gene(cls, gene: Gene) -> HPOSet: ...
    @classmethod
    def from_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_omim_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_orpha_disease(cls, disease: Orpha) -> HPOSet: ...
    def __len__(self) -> int: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> Iterator[HPOTerm]: ...
    def __contains__(self, term: HPOTerm) -> bool: ...


class HPOPhenoSet:
    def __init__(self, terms: List[int | HPOTerm]): ...
    def add(self, term: int | HPOTerm): ...
    def child_nodes(self) -> HPOSet: ...
    def remove_modifier(self) -> HPOSet: ...
    def replace_obsolete(self) -> HPOSet: ...
    def all_genes(self) -> Set[Gene]: ...
    def omim_diseases(self) -> Set[Omim]: ...
    def orpha_diseases(self) -> Set[Orpha]: ...
    def information_content(self) -> Dict[str, Any]: ...
    def similarity(self, other: HPOSet, kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> float: ...
    def similarity_scores(self, other: List[HPOSet], kind: str = "omim", method: str = "graphic", combine: str = "funSimAvg") -> List[float]: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    def serialize(self) -> str: ...
    def to_bytes(self) -> bytes: ...
    def terms(self) -> Iterator[HPOTerm]: ...
    @classmethod
    def from_queries(cls, queries: List[int | str]) -> HPOSet: ...
    @classmethod
    def from_serialized(cls, pickle: str) -> HPOSet: ...
    @classmethod
    def from_bytes(cls, bytes: bytes) -> HPOSet: ...
    @classmethod
    def from_gene(cls, gene: Gene) -> HPOSet: ...
    @classmethod
    def from_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_omim_disease(cls, disease: Omim) -> HPOSet: ...
    @classmethod
    def from_orpha_disease(cls, disease: Orpha) -> HPOSet: ...
    def __len__(self) -> int: ...
    def __str__(self) -> str: ...
    def __iter__(self) -> Iterator[HPOTerm]: ...
    def __contains__(self, term: HPOTerm) -> bool: ...


class Ontology:
    def __init__(self, data_folder: str = "", from_obo_file: bool = True): ...
    # We're documenting the Ontology as if it were a static method,
    # because it is exposed as a Singleton and not as a class
    genes: Collection[Gene]
    omim_diseases: Collection[Omim]
    orpha_diseases: Collection[Orpha]
    @staticmethod
    def get_hpo_object(query: str | int) -> HPOTerm: ...
    @staticmethod
    def match(query: str) -> HPOTerm: ...
    @staticmethod
    def path(query1: str | int, query2: str | int) -> Tuple[int, List[HPOTerm], int, int]: ...
    @staticmethod
    def search(query: str) -> Iterator[HPOTerm]: ...
    @staticmethod
    def hpo(id: int) -> HPOTerm: ...
    @staticmethod
    def version() -> str: ...
    @staticmethod
    def set_custom_information_content(scores: List[Tuple[int, float]]): ...
    @staticmethod
    def __call__(data_folder: str = "", from_obo_file: bool = True, transitive: bool = False): ...
    @staticmethod
    def __len__() -> int: ...
    @staticmethod
    def __repr__() -> str: ...
    @staticmethod
    def __getitem__(id_: int) -> HPOTerm: ...
    @staticmethod
    def __iter__() -> Iterable[HPOTerm]: ...



from pyhpo.pyhpo import EnrichmentModel
from pyhpo.pyhpo import linkage

class HPOEnrichment:
    """
    Not implemented
    """
    def __init__(self, *args, **kwargs):
        pass

    def enrichment(self, *args, **kwargs):
        raise NotImplementedError


__all__ = (
    "EnrichmentModel",
    "linkage",
    "HPOEnrichment",
)



from pyhpo.pyhpo import Ontology
from pyhpo.pyhpo import _Ontology
from pyhpo.pyhpo import Gene
from pyhpo.pyhpo import Omim
from pyhpo.pyhpo import Orpha
from pyhpo.pyhpo import HPOTerm
from pyhpo.pyhpo import HPOSet
from pyhpo.pyhpo import BasicHPOSet
from pyhpo.pyhpo import HPOPhenoSet
from pyhpo.pyhpo import __version__
from pyhpo.pyhpo import __backend__

from pyhpo import annotations
from pyhpo import stats
# import pyhpo.set
from pyhpo import helper

__all__ = (
    "_Ontology",
    "Ontology",
    "Gene",
    "Omim",
    "Orpha",
    "HPOTerm",
    "HPOSet",
    "BasicHPOSet",
    "HPOPhenoSet",
    "__version__",
    "__backend__",
    "annotations",
    "stats",
    "helper",
)



#!/usr/bin/env python3

"""
Checks that the tag-version matches the Python package version in pyproject.toml
and that the version does not yet exist on PyPI
"""


import os
from urllib import request
from urllib.error import HTTPError
import re
import sys
import tomllib
from typing import Optional


def version_from_tag() -> Optional[str]:
    version = os.getenv('GITHUB_REF')
    if version:
        version = re.sub('^refs/tags/v*', '', version.lower())
        return version
    else:
        print("No version tag found")
        return None


def version_from_toml() -> Optional[str]:
    with open("Cargo.toml", "rb") as f:
        data = tomllib.load(f)
        try:
            return data['package']['version']
        except KeyError:
            print("No version defined in pyproject.toml")
            return None


def present_in_pypi(version: str) -> bool:
    url = f"https://pypi.org/pypi/hpo3/{version}/json"
    print(f"Checking {url}")
    try:
        with request.urlopen(url) as response:
            if response.status == 200:
                print(f"Version {version} is already present in PyPi")
                return True
            else:
                print(f"PyPI did not respond correctly for version {version}")
                print(f"Received response status {response.status}")
                return True
    except HTTPError:
        # In theory, the request could fail for many reasons, but most likely
        # it means that the version is not yet present in PyPI
        return False


def main():
    git_tag = version_from_tag()
    if git_tag is None:
        return 1
    toml_version = version_from_toml()
    if toml_version is None:
        return 1
    if git_tag == toml_version:
        if present_in_pypi(git_tag):
            return 1
        else:
            return 0
    else:
        print(f"Different versions in tag ({git_tag}) and toml ({toml_version})")
        return 1


if __name__ == "__main__":
    sys.exit(main())



from pyhpo.pyhpo import batch_similarity
from pyhpo.pyhpo import batch_set_similarity
from pyhpo.pyhpo import batch_gene_enrichment
from pyhpo.pyhpo import batch_disease_enrichment
from pyhpo.pyhpo import batch_omim_disease_enrichment
from pyhpo.pyhpo import batch_orpha_disease_enrichment

__all__ = (
    "batch_similarity",
    "batch_set_similarity",
    "batch_gene_enrichment",
    "batch_disease_enrichment",
    "batch_omim_disease_enrichment",
    "batch_orpha_disease_enrichment",
)



from typing import Any, List, Tuple, TypedDict
from pyhpo import HPOSet, HPOTerm
from pyhpo.annotations import Gene, Omim


class EnrichmentOutput(TypedDict):
    enrichment: float
    fold: float
    count: int
    item: Gene | Omim

class HpoEnrichmentOutput(TypedDict):
    hpo: HPOTerm
    count: int
    enrichment: float

class EnrichmentModel:
    def __init__(self, category: str): ...
    def enrichment(
        self,
        method: str,
        hposet: HPOSet
    ) -> List[EnrichmentOutput]: ...


class HPOEnrichment:
    def __init__(self, category: str): ...
    def enrichment(
        self,
        method: str,
        annotation_sets: List[Omim | Gene]
    ) -> List[HpoEnrichmentOutput]: ...


def linkage(
    sets: List[HPOSet],
    method: str,
    kind: str,
    similarity_method: str,
    combine: str
) -> List[Tuple[int, int, float, int]]: ...



import unittest

from pyhpo import Ontology
from pyhpo.set import HPOSet
from pyhpo.stats import EnrichmentModel
from pyhpo import annotations as an

# Number of terms in HPO Ontology
# grep "^\[Term\]$" pyhpo/data/hp.obo | wc -l
N_TERMS = 19903

# Number of genes in the annotation dataset
# cut -f4 pyhpo/data/phenotype_to_genes.txt | grep -v "^#" | grep -v "^gene_symbol" | sort -u | wc -l  # noqa: E501
# cut -f1 example_data/2024-03-06/genes_to_phenotype.txt | grep -v "^ncbi_gene_id" | sort -u | wc -l
N_GENES = 5178

# Number of OMIM diseases in the annotation dataset
# cut -f1,3 pyhpo/data/phenotype.hpoa | grep "^OMIM" | sort -u | cut -f2 | grep -v "NOT" | wc -l  # noqa: E501
N_OMIM = 8487

# Number of ORPHA diseases in the annotation dataset
# cut -f1,3 pyhpo/data/phenotype.hpoa | grep "^ORPHA" | sort -u | cut -f2 | grep -v "NOT" | wc -l  # noqa: E501
N_ORPHA = 4314


class IntegrationFullTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        Ontology()
        cls.terms = Ontology

    def test_terms_present(self):
        """
        These test will most likely need to be updated
        after every data update
        """
        assert len(self.terms) == N_TERMS, len(self.terms)

    def test_genes_associated(self):
        """
        These test will most likely need to be updated
        after every data update
        """
        assert len(self.terms.genes) == N_GENES, len(self.terms.genes)

    def test_omim_associated(self):
        """
        These test will most likely need to be updated
        after every data update
        """
        self.assertEqual(
            len(self.terms.omim_diseases),
            N_OMIM
        )

    def test_orpha_associated(self):
        """
        These test will most likely need to be updated
        after every data update
        """
        self.assertEqual(len(self.terms.orpha_diseases), N_ORPHA)

    def test_average_annotation_numbers(self):
        """
        These test will most likely need to be updated
        after every data update
        """
        genes = []
        omim = []
        for term in self.terms:
            genes.append(len(term.genes))
            omim.append(len(term.omim_diseases))

        assert sum(genes)/len(genes) > 36, sum(genes)/len(genes)
        assert sum(omim)/len(omim) > 29, sum(omim)/len(omim)

    def test_annotation_inheritance(self):
        for term in self.terms:
            lg = len(term.genes)
            lo = len(term.omim_diseases)

            for child in term.children:
                with self.subTest(t=term.id, c=child.id):
                    assert lg >= len(child.genes)
                    assert child.genes.issubset(term.genes)

                    assert lo >= len(child.omim_diseases)
                    assert child.omim_diseases.issubset(term.omim_diseases)

            for parent in term.parents:
                with self.subTest(t=term.id, c=parent.id):
                    assert lg <= len(parent.genes)
                    assert term.genes.issubset(parent.genes)

                    assert lo <= len(parent.omim_diseases)
                    assert term.omim_diseases.issubset(parent.omim_diseases)


    def test_relationships(self):
        for term in self.terms:
            for child in term.children:
                assert child.child_of(term)
                assert not child.parent_of(term)
                assert term.parent_of(child)
                assert not term.child_of(child)

            for parent in term.parents:
                assert not parent.child_of(term)
                assert parent.parent_of(term)
                assert not term.parent_of(parent)
                assert term.child_of(parent)

    def test_set(self):
        full_set = HPOSet.from_queries(
            [int(x) for x in self.terms]
        )

        self.assertEqual(
            len(full_set),
            len(self.terms)
        )

        phenoterms = full_set.remove_modifier()
        self.assertLess(
            len(phenoterms),
            len(full_set)
        )
        self.assertGreater(
            len(phenoterms),
            0
        )

        self.assertIn(
            self.terms[5],
            full_set
        )

        self.assertNotIn(
            self.terms[5],
            phenoterms
        )

    def test_gene_enrichment(self):
        hposet = HPOSet.from_queries('HP:0007401,HP:0010885'.split(','))
        gene_model = EnrichmentModel('gene')
        res = gene_model.enrichment('hypergeom', hposet)
        self.assertIsInstance(res, list)
        self.assertIn('item', res[0])
        self.assertIn('count', res[0])
        self.assertIn('enrichment', res[0])
        self.assertIsInstance(res[0]['item'], an.Gene)
        self.assertIsInstance(res[0]['count'], int)
        self.assertIsInstance(res[0]['enrichment'], float)

    def test_omim_enrichment(self):
        hposet = HPOSet.from_queries('HP:0007401,HP:0010885'.split(','))
        omim_model = EnrichmentModel('omim')
        res = omim_model.enrichment('hypergeom', hposet)
        self.assertIsInstance(res, list)
        self.assertIn('item', res[0])
        self.assertIn('count', res[0])
        self.assertIn('enrichment', res[0])
        self.assertIsInstance(res[0]['item'], an.Omim)
        self.assertIsInstance(res[0]['count'], int)
        self.assertIsInstance(res[0]['enrichment'], float)

    def test_orpha_enrichment(self):
        hposet = HPOSet.from_queries("HP:0007401,HP:0010885".split(","))
        orpha_model = EnrichmentModel('orpha')
        res = orpha_model.enrichment("hypergeom", hposet)
        self.assertIsInstance(res, list)
        self.assertIn("item", res[0])
        self.assertIn("count", res[0])
        self.assertIn("enrichment", res[0])
        self.assertIsInstance(res[0]["item"], an.Orpha)
        self.assertIsInstance(res[0]["count"], int)
        self.assertIsInstance(res[0]["enrichment"], float)

    def test_custom_information_scores(self):
        term1 = Ontology.hpo(10885)
        self.assertEqual(term1.information_content.custom, 0.0)

        term2 = Ontology.hpo(7401)
        self.assertEqual(term1.information_content.custom, 0.0)

        Ontology.set_custom_information_content([(10885, 1.234)])
        self.assertAlmostEqual(term1.information_content.custom, 1.234)
        self.assertAlmostEqual(term2.information_content.custom, 0.0)



from typing import Any, Dict, Set
from pyhpo.pyhpo import HPOSet

class Gene:
    id: int
    name: str
    hpo: Set[int]
    def hpo_set(self) -> HPOSet: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    @classmethod
    def get(cls, query: int|str) -> 'Gene': ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __hash__(self) -> int: ...


class Omim: 
    id: int
    name: str
    hpo: Set[int]
    def hpo_set(self) -> HPOSet: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    @classmethod
    def get(cls, query: int|str) -> 'Omim': ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __hash__(self) -> int: ...

class Orpha:
    id: int
    name: str
    hpo: Set[int]
    def hpo_set(self) -> HPOSet: ...
    def toJSON(self, verbose: bool = False) -> Dict[str, Any]: ...
    @classmethod
    def get(cls, query: int|str) -> 'Orpha': ...
    def __str__(self) -> str: ...
    def __int__(self) -> int: ...
    def __hash__(self) -> int: ...



# PYTHONPATH=./ python tests/benchmarking.py
import cProfile

from pyhpo import Ontology
from pyhpo import BasicHPOSet
from pyhpo.annotations import Omim


def diseases2basicsets(diseases):
    return [(
        d.name,
        BasicHPOSet.from_queries(list(d.hpo))
    ) for d in diseases]


def build_ontology():
    _ = Ontology()


def compare_set(diseases):
    gaucher = Omim.get(230800)
    set1 = BasicHPOSet.from_queries(list(gaucher.hpo))
    for x in diseases:
        set2 = BasicHPOSet.from_queries(list(x.hpo))
        _ = set1.similarity(set2)
    return None


print('===== LOADING ONTOLOGY ======')
cProfile.run('build_ontology()')

print('===== BUILDING DISEASE BasicSETS ======')
cProfile.run('diseases2basicsets(Ontology.omim_diseases)')

print('===== BUILDING GENE BasicSETS ======')
cProfile.run('diseases2basicsets(Ontology.genes)')

print('===== COMPARING SETS ======')
cProfile.run('compare_set(Ontology.omim_diseases)')





from typing import Any, Dict, List, Tuple
from pyhpo.pyhpo import HPOSet

from pyhpo.pyhpo import HPOTerm


def batch_similarity(
    comparisons: List[Tuple[HPOTerm, HPOTerm]],
    kind:str,
    method: str
) -> List[float]: ...
def batch_set_similarity(
    comparisons: List[Tuple[HPOSet, HPOSet]],
    kind:str,
    method: str,
    combine: str
) -> List[float]: ...
def batch_gene_enrichment(hposets: List[HPOSet]) -> List[List[Dict[str, Any]]]: ...
def batch_disease_enrichment(hposets: List[HPOSet]) -> List[List[Dict[str, Any]]]: ...
def batch_omim_disease_enrichment(hposets: List[HPOSet]) -> List[List[Dict[str, Any]]]: ...
def batch_orpha_disease_enrichment(hposets: List[HPOSet]) -> List[List[Dict[str, Any]]]: ...